/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AdBlock for Mozilla.
 *
 * The Initial Developer of the Original Code is
 * Henrik Aasted Sorensen.
 * Portions created by the Initial Developer are Copyright (C) 2002
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * Henrik Aasted Sorensen <henrik@aasted.org>
 * Stefan Kinitz <mcmurmel.blah@gmx.de>
 * Rue <quill@ethereal.net>
 *
 * AdBlockLearner additions:
 * Babak Esfandiari <babak@sce.carleton.ca>
 * Richard Nock provided the theory.
 *
 * ***** END LICENSE BLOCK ***** */


var Branch;

// Array holding all the patterns to block.
var Patterns;
// Boolean signalling whether to remove or hide images. Images will be hidden if true. They will be removed if false.
var Hide;
// These tags are being processed.
var tags = ["img", "embed", "object", "iframe"]; // iframe is still listed so it can be killed at its source, if desired


// AdBlockLearnerTest variables
var maxSteps;
var stepIncrement;
var total; //total examples
var totalexamples; 
var step;  //each step, we save a frozen version of the current items and weights
var confusionMatrix = new Array(maxSteps); //each row is a confusion matrix: one per step 
var confusionMatrixTable; //0(a): Neg/neg; 1(b): Neg/pos; 2(c): Pos/neg; 3(d): Pos/pos
var visits = 0;

//
// AdBlockLearner variables
// 

var Items = new Array(maxSteps);
var NegItems = new Array(maxSteps);
var Weights = new Array(maxSteps);
var NegWeights = new Array(maxSteps);
var ItemsTable;
var NegItemsTable;
var WeightsTable;
//var WeightConstant;
var NegWeightsTable;
var prefObj;
var BlockedItems; 
var NonBlockedItems;
//var beta = 2.73;
var beta = 0.61;
//var updateAll = true;



// Loads the preferences. 
function loadSettings() {

	//alert("Loading.......");
	try {
		prefObj = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService);
		Branch = prefObj.getBranch("adblock.");
		Enabled = Branch.getBoolPref("enabled");
		Hide = Branch.getBoolPref("hide");
		Patterns = Branch.getCharPref("patterns").split(" ");

		//AdBlockLearner additions
		maxSteps = Branch.getCharPref("maxSteps") * 1.0;
		stepIncrement = Branch.getCharPref("stepIncrement") * 1.0;

		step = Branch.getCharPref("step");
		total = Branch.getCharPref("total") * 1.0;
		totalexamples = Branch.getCharPref("totalexamples") * 1.0;
		visits = Branch.getCharPref("visits");
		for (var i = 0 ; i <= step ; i++) {
			Items[i] = Branch.getCharPref("items" + i);
			NegItems[i] = Branch.getCharPref("negitems" + i);
			Weights[i] = Branch.getCharPref("weights" + i);
			//WeightConstant[i] = Branch.getCharPref("constant" + i);
			NegWeights[i] = Branch.getCharPref("negweights" + i);
			confusionMatrix[i] = Branch.getCharPref("confusionMatrix" + i);
		}

	} catch(exception) {}

	visits++;

	// If AdBlock is not enabled... Return false as a signal that the program should not filter.
	if (!Enabled) {
		return false;
	}
		
	for (var i = 0 ; i < Patterns.length ; i++) {
		Patterns[i] = new RegExp(convert2RegExp(Patterns[i]), "i");
	}
	return true;
}
 
// Resolves the absolute URL of the image-address being sent as an argument.
// page is the document which contains the image.
function getImageURL(imageSrc, page) {
	var URL = Components.classes["@mozilla.org/network/standard-url;1"].createInstance(Components.interfaces.nsIURL);	
	URL.spec = page.URL;
	return URL.resolve(imageSrc);	
}

// This method checks if a picture matches a filter.
// It returns true if the imageurl is caught by one of the filters.
// It returns false if the imageurl is not caught by any filter.
// adblocklearnertest note: we're cannibalizing this method for instrumentation purposes

function validate(imageurl) {

	//alert("Validating " + imageurl);

	var expert;
	var isValid = false;

	for (var i = 0 ; ((i < Patterns.length) && (!isValid)) ; i++) {
		if (Patterns[i].test(imageurl)) isValid = true;
	}

	// first update frozen experts' confusion matrices
	for (var j = 0 ; j < step ; j++) {
		expert = expertvalidate(imageurl, j);
		confusionMatrixTable = confusionMatrix[j].split(" ");
		
		if ((!isValid) && (!expert)) confusionMatrixTable[0]++;
		if ((!isValid) && (expert)) confusionMatrixTable[1]++;
		if ((isValid) && (!expert)) confusionMatrixTable[2]++;
		if ((isValid) && (expert)) confusionMatrixTable[3]++;

		confusionMatrix[j] = confusionMatrixTable.join(" ");
	}

	// now update current experts
	expert = expertvalidate(imageurl, step);

	// alert("expert: " + expert + "; valid: " + isValid);


	if ((!expert) && (isValid)) learn(imageurl); //learned on disagreement!	
	if ((expert) && (!isValid)) unlearn(imageurl); //unlearn on disagreement!
	
	//increment total
	//if reached the increment, update the step and copy over the next experts

	total++;

	//alert("incremented total" + total);

	if ((total % stepIncrement) == 0) {
		confusionMatrix[step] = "0 0 0 0"; //confusion matrix for latest frozen step
		step++;
		confusionMatrix[step] = "0 0 0 0"; //confusion matrix for new step (not used)

		//copy latest experts over to the new step
		Items[step] = Items[step-1];
		NegItems[step] = NegItems[step-1];
		Weights[step] = Weights[step-1];
		//WeightConstant[step] = WeightConstant[step-1];
		NegWeights[step] = NegWeights[step-1];
	}

	
	//savePrefs();
	return isValid;
}

// same as validate, but according to the learned experts

function expertvalidate(imageurl, index) {

	//alert("Expert validating " + imageurl);

	ItemsTable = Items[index].split(" ");
	WeightsTable = Weights[index].split(" ");
	NegItemsTable = NegItems[index].split(" ");
	NegWeightsTable = NegWeights[index].split(" ");
			
	var sum = 0;
	for (var i = 0 ; i < ItemsTable.length ; i++) {
		if ( imageurl.indexOf(ItemsTable[i]) != -1) {
			//alert(ItemsTable[i] + " matches " + imageurl);
			sum += (1 * WeightsTable[i]);
		}
	}

	for (var i = 0 ; i < NegItemsTable.length ; i++) {
		if ( imageurl.indexOf(NegItemsTable[i]) != -1) {
			//alert(NegItemsTable[i] + " matches " + imageurl);
			sum -= (1 * NegWeightsTable[i]);
		}
	}

	//alert("End of Expert validating " + imageurl);


	//if (sum >= WeightConstant) return true;	
	if (sum > 0) return true;
	return false;
}



function learn(imgUrl) {

	//alert("Learning " + imgUrl);

	totalexamples++;
	//first let's update the weights

	for (var i = 0; i < ItemsTable.length; i++) {
		if ( imgUrl.indexOf(ItemsTable[i]) != -1) {
			//it matches: promote the expert
			WeightsTable[i] = (1/beta) * WeightsTable[i];
		}
		else {
			// the coward abstained! punish him mildly
			WeightsTable[i] = ( (1 + (beta * beta))/ (2 * beta) ) * WeightsTable[i];
		}
	}

	for (var i = 0; i < NegItemsTable.length; i++) {
		if ( imgUrl.indexOf(NegItemsTable[i]) != -1) {
			//it matches: punish the expert
			NegWeightsTable[i] = beta * NegWeightsTable[i];
		}
		else {
			// the coward abstained! punish him mildly
			NegWeightsTable[i] = ( (1 + (beta * beta))/ (2 * beta) ) * NegWeightsTable[i];
		}
	}

	Weights[step] = WeightsTable.join(" ");
	NegWeights[step] = NegWeightsTable.join(" ");

	//tokenize the URL, get rid of "http://" (the first two tokens) and the filename
	tokens = imgUrl.split("/").slice(2, -1);
	
	//now possibly add new tokens to the expert list

	for (var i = 0 ; i < tokens.length ; i++) {
		var found = false;
		for (var j = 0; ((j < ItemsTable.length) && (found == false)) ; j++) {
			if (tokens[i] == ItemsTable[j]) {
				found = true;
			}
		}
		//only add tokens that aren't already in the items list
		if (found == false) {
			//update items... 
			Items[step] += (" " + tokens[i]);
			Weights[step] += (" " + Math.pow( ((1 + (beta * beta))/ (2 * beta) ), totalexamples-1) );
		}

	}

	//alert("Learned: " + Items[step]);


	//update tables
	//ItemsTable = Items.split(" ");
	//WeightsTable = Weights.split(" ");

	savePrefs ();
}

function unlearn(unblockedItem) {

	totalexamples++;
	//alert("Unlearning " + unblockedItem);

	for (var i = 0; i < ItemsTable.length; i++) {
		if ( unblockedItem.indexOf(ItemsTable[i]) != -1) {
			//it matches: punish the expert
			WeightsTable[i] = beta * WeightsTable[i];
		}
		else {
			// the coward abstained! punish him mildly
			WeightsTable[i] = ( (1 + (beta * beta))/ (2 * beta) ) * WeightsTable[i];
		}
	}

	for (var i = 0; i < NegItemsTable.length; i++) {
		if ( unblockedItem.indexOf(NegItemsTable[i]) != -1) {
			//it matches: promote the expert
			NegWeightsTable[i] = (1/beta) * NegWeightsTable[i];
		}
		else {
			// the coward abstained! punish him mildly
			NegWeightsTable[i] = ( (1 + (beta * beta))/ (2 * beta) ) * NegWeightsTable[i];
		}
	}

	Weights[step] = WeightsTable.join(" ");
	NegWeights[step] = NegWeightsTable.join(" ");

	tokens = unblockedItem.split("/").slice(2, -1);

	for (var i = 0 ; i < tokens.length ; i++) {
		var found = false;
		for (var j = 0; ((j < NegItemsTable.length) && (found == false)) ; j++) {
			if (tokens[i] == NegItemsTable[j]) {
				found = true;
			}
		}
		//only add tokens that aren't already in the items list
		if (found == false) {
			//update items... 
			NegItems[step] += (" " + tokens[i]);
			NegWeights[step] += (" " + Math.pow( ((1 + (beta * beta))/ (2 * beta) ), totalexamples-1) );
		}

	}

	//update tables
	//NegItemsTable = NegItems.split(" ");
	//NegWeightsTable = NegWeights.split(" ");

	//alert("Updated Weights: " + Weights + NegWeights);

	//alert("Unlearned: " + NegItems[step]);

	savePrefs();
}




// Converts a pattern in this programs simple notation to a regular expression.
function convert2RegExp( pattern ) {
	// If the user already entered a regular expression.
	if ( isRegExp( pattern ) ) {
		return pattern.substr(1, pattern.length - 2);
	}
	
	s = new String(pattern);
	res = new String("^");
	
	for (var i = 0 ; i < s.length ; i++) {
		switch(s[i]) {
			case '*' : 
				res += ".*";
				break;
				
			case '.' : 
			case '?' :
			case '^' : 
			case '$' : 
			case '+' :
			case '{' :
			case '[' : 
			case '|' :
			case '(' : 
			case ')' :
			case ']' :
				res += "\\" + s[i];
				break;
			
			case '\\' :
				res += "\\\\";
				break;
			
			case ' ' :
				// Remove spaces from URLs.
				break;
			
			default :			
				res += s[i];
				break;
		}
	}
	return res + '$'	;
}


// initiates filtering on the given document (if possible).
//    --accepts either the target-document directly, or through its parent-event
//      --(mozilla sends an event parameter by default if nothing was directly passed *and* there's an event trigger)

function filter(eventorotherparam) {
	// If loadsettings returns false, it means that the page should not be filtered.				
	if (!loadSettings()) {
		return;
	}

	// if we were called by an event -- (as opposed to another function)
	if (eventorotherparam.type) {
		if (eventorotherparam.target.contentDocument)	
			var page = eventorotherparam.target.contentDocument; // This is the tab. We need to get the contents.		
	} else
		var page = eventorotherparam; // we were sent the document directly -- (by the function)

	parseFrames(page);	// initiate the filtering!
}

// parses document looking for frames, *recursively* -- (embedded frames?)
function parseFrames(page) {
	// if the element/page has frames
	if (page.getElementsByTagName("frame").length > 0) {		
		for (var f = 0 ; f < page.getElementsByTagName("frame").length ; f++) {
			var frame = page.getElementsByTagName("frame").item(f).contentDocument;
			parseFrames(frame);
		}
  	}	
	else 
		parseInlineFrames(page);
}

// parses document looking for iframes -- *Inline-Frames*
function parseInlineFrames(page) {	
	// if the element/page has iframes
	if (page.getElementsByTagName("iframe").length > 0) {
		for (var f = 0 ; f < page.getElementsByTagName("iframe").length ; f++) {
			var iframe = page.getElementsByTagName("iframe").item(f).contentDocument;
			parsePage(iframe);
		}
  	}
	// even if there weren't -- "onward, ho -!"
	parsePage(page);
}

// Goes through all relevant tags and determines if they should stay or be banished. :)
function parsePage(page) {
	for (var j = 0 ; j < tags.length ; j++) {
		var scanTags = page.getElementsByTagName( tags[j] );

		for (var i = 0 ; i < scanTags.length ; i++) {
			var src = findAttribute(scanTags[i],"src");
			if (src && validate( getImageURL(src, page))) {
				if (Hide) {
					scanTags[i].style.visibility = "hidden";
				} else {
					scanTags[i].style.display = "none";
				}
			}
		}
	}
}
	
// Checks an element and its children recursively for an attribute.
function findAttribute(element, name) {
	if (element.hasAttribute(name)) {
		return element.getAttribute(name);
	}
	
	if (element.hasChildNodes()) {
		for (var i = 0 ; i < element.childNodes.length ; i++) {
			var res = findAttribute(element.childNodes[i], name);
			if (res != false) 
				return res;
		}
	}
	
	return false; // Didn't find it! 
}

// Open the settings window.
function settings() {
	window.open("chrome://adblock/content/settings.xul", "AdBlock Preferences", "chrome,resizable,centerscreen");
}

function makemenu() {
  if(gContextMenu) {  		
  		var menuitem = document.getElementById("adblock-menuitem");  		
        if (gContextMenu.onImage) {
        	menuitem.hidden = false;        	
        } else {
        	menuitem.hidden = true;
        }
    }
}

function addFilterDialog() {
	var page = document.getElementById("content").contentDocument;

	// passes the current page to the dialogue via window.argument[1] -- allows the page to be refiltered on 'accept'
	window.openDialog("chrome://adblock/content/addfilterdialog.xul","Add filter", "chrome,modal,centerscreen", getImageURL(gContextMenu.target.getAttribute("src"),page), page);
}


//called whenever learning happens
function savePrefs () {
	//window.alert("Saving prefs!");
	Branch.setCharPref("step", step);
	Branch.setCharPref("total", total);
	Branch.setCharPref("totalexamples", totalexamples);
	Branch.setCharPref("visits", visits);
	for (var i = 0 ; i <= step ; i++) {

		//window.alert("Saving items until step " + step + ": " + Items[i]);

		Branch.setCharPref("items" + i, Items[i]);
		Branch.setCharPref("weights" + i, Weights[i]);
		//Branch.setCharPref("constant" + i, WeightConstant[i]);
		Branch.setCharPref("negitems" + i, NegItems[i]);
		Branch.setCharPref("negweights" + i, NegWeights[i]);
		Branch.setCharPref("confusionMatrix" + i, confusionMatrix[i]);
	}
}

function saveAndPrompt () {
	savePrefs();
	// are we done yet?
	if (step >= maxSteps) {
		alert("Congratulations! You have finished testing AdBlockLearner. Please click OK and exit the browser, then copy the environment variables in Prefs.js that start with the adblock prefix into a text file and email it to babak@sce.carleton.ca . If you are concerned about your privacy, you can simply copy the 10 variables called confusionMatrix, and leave the rest out.");
	}
}
	
	
function initAdBlock() {

	// Initiate settings:	
	prefObj = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService);
	Branch = prefObj.getBranch("adblock.");
	if (! Branch.prefHasUserValue("patterns")) {
		Branch.setCharPref("patterns", "");	
		Branch.setBoolPref("enabled", true);	
		Branch.setBoolPref("hide", false);	
	}
	//AdBlockLearner additions
	if (! Branch.prefHasUserValue("items0")) {
            Branch.setCharPref("step", "0");
		Branch.setCharPref("total", "0");
		Branch.setCharPref("totalexamples", "0");
		Branch.setCharPref("visits", "0");
		Branch.setCharPref("items0", "http:");
		Branch.setCharPref("weights0", "0");
		//Branch.setCharPref("constant0", "1");
		Branch.setCharPref("negitems0", "http:");
		Branch.setCharPref("negweights0", "0");
		Branch.setCharPref("confusionMatrix0", "0 0 0 0");
	}

}





// DEBUG-functions
// ---------------
function debugmsg(str) {
	window.defaultStatus=str;
}


// lists everything in an object -- Useful debug-function.
function listObject(obj, s) {
	var res = "List: " + obj + "\n";
	for(var list in obj) {		
		if (list.indexOf(s) != -1)
			res += list + ", ";
	}
		
	alert(res);
}
// lists everything in an object -- Useful debug-function.
function invListObject(obj, s) {
	var res = "List: " + obj + "\n";
	for(var list in obj) {		
		if (list.indexOf(s) == -1)
			res += list + ", ";
	}
		
	alert(res);
}

