/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AdBlock for Mozilla.
 *
 * The Initial Developer of the Original Code is
 * Henrik Aasted Sorensen.
 * Portions created by the Initial Developer are Copyright (C) 2002
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * Henrik Aasted Sorensen <henrik@aasted.org>
 * Stefan Kinitz <mcmurmel.blah@gmx.de>
 * Rue <quill@ethereal.net>
 * ***** END LICENSE BLOCK ***** */

// Loads the preferences and enters them into the UI.
function loadSettings() {
	// Array holding all the patterns to block.
	var Patterns;
	// Boolean value telling whether AdBlock is enabled or not.
	var Enabled = true;
	// Boolean signalling whether to remove or hide images. Images will be hidden if true. They will be removed if false.
	var Hide = true;

	try {
		prefObj = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService);
		var Branch = prefObj.getBranch("adblock.");
		Enabled = Branch.getBoolPref("enabled");
		Hide = Branch.getBoolPref("hide");
		Patterns = Branch.getCharPref("patterns").split(" ");
	} catch(exception) {}
		
	// Add loaded patterns to the list.
	if (Patterns != "") {
		fillList(Patterns);
	}
	
	var checkbox = document.getElementById("enabled");
	checkbox.checked = Enabled;	
	
	var radiohide = document.getElementById("radio-hide");
	var radioremove = document.getElementById("radio-remove");
	
	if (Hide)
		radiohide.setAttribute("selected", "true");
	else
		radioremove.setAttribute("selected", "true");
}

// Add a filter to the list -- now constructed for inline-editing :)
function addFilter() {
	var newItem;
	var list = document.getElementById("list");
	var textbox = document.getElementById("newfilter");
	var filter = makePattern(textbox.value); // must be made into pattern
	var newItemindex;
	
	// if there's an actual entry
	if (textbox.value != "") {
		// if it's not a regular expression, or it is, BUT we've said 'ok' to the warning
		if (!isRegExp(textbox.value) || (isRegExp(textbox.value) && warnRegExp())) {
			newItem = createListItem(filter);
			textbox.value = "";
			newItemindex = list.getIndexOfItem(newItem);
			list.selectedIndex = newItemindex; // sets the list's selection to the new item
			list.currentItem = newItem;
			list.ensureIndexIsVisible(newItemindex); // scrolls the list to show the new item
		}		
	}
	if (!textbox.focused) 
		textbox.focus(); // always return focus to the textbox
}

function fillList(Patterns) {	
	var list = document.getElementById("list");
	while (list.getRowCount() != 0)
		list.removeItemAt(0);
		
	// Sort the patterns
	Patterns.sort();	
	
	// fill the list with patterns
	for (var i = 0 ; i < Patterns.length; i++) {
		createListItem(Patterns[i]);
	}
}

// Asks the user if he really wants to clear the list.
function clearList() {
	if (confirm("Do you really want to clear the list?")) {
		var list = document.getElementById("list");
		while (list.getRowCount() > 0)
			list.removeItemAt(0);
	}
}

// Imports filters from disc.
function importList() {
	var fp = Components.classes["@mozilla.org/filepicker;1"].createInstance(Components.interfaces.nsIFilePicker);
	var stream = Components.classes["@mozilla.org/network/file-input-stream;1"].createInstance(Components.interfaces.nsIFileInputStream);
	var streamIO = Components.classes["@mozilla.org/scriptableinputstream;1"].createInstance(Components.interfaces.nsIScriptableInputStream);

	fp.init(window, "Select a File", fp.modeOpen);
	fp.appendFilters(fp.filterText);

	if (fp.show() != fp.returnCancel) {
		stream.init(fp.file, 0x01, 0444, null);
		streamIO.init(stream);
		var input = streamIO.read(stream.available());
		streamIO.close();
		stream.close();
		input = input.split("\n");
		if (input[0] == "[AdBlock]") {
			input.shift();
			
			Patterns = getPatterns();
			if (Patterns != "") // To avoid an empty entry in the list after importing.
				Patterns = Patterns.split(" ");						
			else 
				Patterns = new Array();
				
			Patterns = Patterns.concat(input);
			
			fillList(Patterns);
		}
		else 
			alert("File not valid.");
	}
}

// Exports the current list of filters to a file on disc.
function exportList() {
	var fp = Components.classes["@mozilla.org/filepicker;1"].createInstance(Components.interfaces.nsIFilePicker);
	var stream = Components.classes["@mozilla.org/network/file-output-stream;1"].createInstance(Components.interfaces.nsIFileOutputStream);

	fp.init(window, "Select a File", fp.modeSave);
	fp.appendFilters(fp.filterText);

	if (fp.show() != fp.returnCancel) {
		if (fp.file.exists())
			fp.file.remove(true);
		fp.file.create(fp.file.NORMAL_FILE_TYPE, 0666);

		stream.init(fp.file, 0x02, 0x200, null);
		stream.write("[AdBlock]\n", 10);
		
		Patterns = getPatterns().split(" ");
		
 		for (var i = 0; i < Patterns.length ; i++) {
			stream.write(Patterns[i], Patterns[i].length);
			if (i < (Patterns.length - 1))
				stream.write("\n", 1);
		}
		stream.close();
	}
}

// Creates a listitem with the proper children for inline-editing -- returns the item's node -- label is set to 'value'
function createListItem(value) {
	var list = document.getElementById("list");
	var templateListItem = document.getElementById("li-template");
	var newItem = list.appendChild(templateListItem.cloneNode(true));
	var newItemcell;
	var newItemfield;
	
	newItem.setAttribute("creating", "true");
	newItemcell = findElementWithParentAttribute("listcell", "creating", "true")
	newItemfield = findElementWithParentAttribute("textbox", "creating", "true")
	newItem.removeAttribute("creating");
	
	newItem.setAttribute("hidden", "false");
	newItem.setAttribute("id", "li-generated");
	newItem.setAttribute("label", value); // sets the listItem's label for later saving -- needed by saveSettings() -- deactivateModify does this too.
	newItem.setAttribute("ondblclick", "listClickControl();");
	newItemcell.setAttribute("label", value); // the visible label for our listbox
	newItemcell.setAttribute("id", "li-lc-generated");
	newItemfield.setAttribute("id", "li-tb-generated");
	
	list.appendChild(newItem);
	return newItem;
}

// This function removes illegal characters (at the moment, space) from the patterns.
function makePattern(pat) {
	var res = "";
	for (var i = 0 ; i < pat.length ; i++) {
		switch (pat[i]) {
			case ' ' : 
				break;
			default :
				res += pat[i];
				break;
		}
	}
	return res;
}

// receives keypresses from the list-area... and makes decisions :)
function listKeyControl(eventX) {

	var keyCodeX = eventX.keyCode; // action-key or '0'
	var charCodeX = eventX.charCode; // character-key or '0'
	keyCodeX += charCodeX; // since one of them is always zero, this works
	
	var list = document.getElementById("list");
	var selected = list.selectedItem;
	var modifying = findElementWithAttribute("listitem", "modifying", "true");
	
	switch (keyCodeX) {		
		case 8:
		case 46: 
			// called by 'delete'
			if (!modifying) // it's possible the user will click another item, but hit 'return' fast enough to be editing the previous one -- the selection index wasn't updated before the keypress
				removeFilter(); 
			break;
			
		case 13:
			// called by 'return'
			if (modifying)
				deactivateModify("return");
			else 
				if (selected) // if there's anything selected
					activateModify();
			break;
				
		case 27:
			// called by 'esc'
			if (modifying)
				deactivateModify("esc"); // sends the 'esc' parameter, which signals to keep the old value
			break;
	}
}

// receives keyClicks from the list-area... and makes decisions :)
function listClickControl() {
	var list = document.getElementById("list");
	var selected = list.selectedItem;
	var modifying = findElementWithAttribute("listitem", "modifying", "true");
	
	// if we have a filter selected and in-focus
	if (selected && selected.focus) {
		// if we're not modifying anything already
		if (!modifying) {
			activateModify();
		}
		// we're already modifying something
		else {
			// if we want to activate a different item
			if (modifying != selected) {
				deactivateModify();
				activateModify();
			}
		}
	}
	// the listbox lost focus
	else {
		if (modifying &! modifying.focused) {
			deactivateModify();
		}
	}
}

// activates the textbox for filter modification -- called by modifyFilter
function activateModify() {
	var list = document.getElementById("list");
	
	if (list.selectedItem) {
		var selectedX = list.selectedItem;
		var selectedXcell = findElementWithParentAttribute("listcell", "selected", "true");
		var selectedXfield = findElementWithParentAttribute("textbox", "selected", "true");
		var filter = selectedXcell.getAttribute("label");
		
		
		selectedX.setAttribute("modifying", "true");		
		selectedXcell.hidden = "true";
		selectedXfield.value = filter; // needs to be called by method, not property
		selectedXfield.setAttribute("hidden", "false");
		selectedX.removeAttribute("ondblclick");
		
		selectedXfield.focus();
		selectedXfield.setAttribute("onblur", "deactivateModify(event.target);"); // <-- *Needs to come AFTER the "list.blur()"*
	}
}

// deactivates the textbox for filter modification -- called by modifyFilter
function deactivateModify(parameter) {
	var modifying = findElementWithAttribute("listitem", "modifying", "true");
	
	if (modifying) {
		var list = document.getElementById("list");
		var selectedIndex = list.selectedIndex;
		var selected = list.selectedItem;
		var disableX = modifying;
		var disableXcell = findElementWithParentAttribute("listcell", "modifying", "true");
		var disableXfield = findElementWithParentAttribute("textbox", "modifying", "true");
		var disableXindex = list.getIndexOfItem(disableX);
		var filter = makePattern(disableXfield.value); // note:  'value' needs to be called by method, not property
		
		// if the filter isn't blank, *And* we weren't called by the "esc" key
		if (filter != "" && parameter != "esc") { 
			disableXcell.setAttribute("label", filter);	// save the filter
			disableX.setAttribute("label", filter); // listitem's label: used for later saving -- needed by saveSettings() -- createListItem does this too.
		}
		
		disableXcell.setAttribute("hidden", "false");
		disableXfield.setAttribute("hidden", "true");
		disableXfield.removeAttribute("onblur");
		disableX.setAttribute("ondblclick", "listClickControl();");
		
		list.selectedIndex = disableXindex; // restore selection
		list.currentItem = list.getItemAtIndex(list.selectedIndex); // and make it the current item
		list.focus(); // refocus the list -- doesn't interfere with refocusing on other controls, but does refocus the *Window* if we were called by losing it to another

		 // keep this for last, so listKeyControl wont accidentally activate the wrong field
		 //    --(if the user deselects and then presses 'return' fast enough)
		disableX.setAttribute("modifying", "false");	}
}

// Removes the selected entry from the list and sets selection to the next item
function removeFilter() {
	var list = document.getElementById("list");
	var textbox = document.getElementById("newfilter");

	// if we have an item to remove
	if (list.selectedItem) {
		var item = list.selectedItem;
		var itemindex = list.getIndexOfItem(item); // the current item's index
		var itemplace = itemindex + 1; // this avoids negative numbers later -- 'itemindex' is zero-based

		var nextitemindex = itemindex; // points to the next item -- equal, since the item below will move up the index once our current item is removed
		var previtemindex = nextitemindex - 1; // points to the previous -- in case we deleted the lowest entry

		var originaltotal = list.getRowCount(); // number of items in list, originally
		var currenttotal = originaltotal-1; // number of items after deletion
		var itemsfollowing = originaltotal - itemplace; // number of items following the deleted item
		
		list.removeChild(item);
		
		// if we have any items left to select
		if (currenttotal > 0) {
			//if there's any items following our deleted entry
			if (itemsfollowing > 0 )
				list.selectedIndex = nextitemindex; // sets the list-selection to the next item
			else
				list.selectedIndex = previtemindex; // sets the list-selection to the previous item
				
			list.ensureIndexIsVisible(list.selectedIndex); // scrolls the list to show the item
		}	
	}
}

// Checks all elements of the specified tagname for an attribute match ON THEIR PARENT --returns first match
function findElementWithParentAttribute(elementTagName, attributeX, valueX) {
	var elementarrayX = document.getElementsByTagName(elementTagName);
	
	for (var z = 0 ; z < elementarrayX.length ; z++) {
		var element = elementarrayX[z];
		if (element.parentNode) {
			if (element.parentNode.hasAttribute(attributeX)) {
				if (element.parentNode.getAttribute(attributeX) == valueX) {
					return element;
				}
			}
		}
	}
	return false; // failure.
}

// Checks all elements of the specified tagname for an attribute match -- returns first match
function findElementWithAttribute(elementTagName, attributeX, valueX) {
	var elementarrayX = document.getElementsByTagName(elementTagName);
	
	for (var z = 0 ; z < elementarrayX.length ; z++) {
		var element = elementarrayX[z];
		if (element.hasAttribute(attributeX)) {
			if (element.getAttribute(attributeX) == valueX) {
				return element;
			}
		}
	}
	return false; // failure.
}

// reselects the currentItem, since the list deselects it on scroll 
//    -- this could just be my old build of mozilla, trippin'...
//		-- "dang" - the hbox also calls at each keypress when an inline-edit is active: now checks 'modifying'
function reSelectListItem () {
	var list = document.getElementById("list");
	var modifying = findElementWithAttribute("listitem", "modifying", "true");
	
	// if we're not modifying anything
	if (!modifying)
		list.selectedIndex = list.getIndexOfItem(list.currentItem); // reselect our current item
}

// persists the list's selected-Index by saving a "backup" copy
//    --because the normal "persist-attribute" route isn't working for this one
function persistListIndex() {
	var list = document.getElementById("list");
	var selectedX = list.selectedIndex;
	
	list.setAttribute("persist-selectedIndex", selectedX);
}

// restores the list's selected-index from the persisted "backup"
function restoreListIndex() {
	var list = document.getElementById("list");
	var textbox = document.getElementById("newfilter");
	var savedIndex = list.getAttribute("persist-selectedIndex");

	// if the list is empty
	if (list.getRowCount() == 0) {
		textbox.focus(); // throw focus to the input-line
	}
	// otherwise, select the last-saved index -- give it focus
	else {
		// if we have a valid index
		if (savedIndex && savedIndex != -1) {
			list.focus();
			//list.ensureIndexIsVisible(savedIndex); // jumps to the given index
			list.scrollToIndex(savedIndex); // scrolls the list to show the index -- more impressive :)
		}
		else {
			list.focus();
			list.ensureIndexIsVisible(list.getRowCount() - 1); // scrolls to the last item -- -1 to make zero-based
		}

		list.selectedIndex = savedIndex; // selects the item we just scrolled to
		//list.focus();
	}
}

// invoked from the settings window -- (moved from adblock.js)
function newFilterDialog() {
	window.openDialog("chrome://adblock/content/addfilterdialog.xul","Add filter","chrome,modal,centerscreen", "");
}

// Returns the patterns in one, long string.
// Use split(" ") to turn it into an array.
function getPatterns() {
	Patterns = "";
	var list = document.getElementById("list");
	
	for (var i = 0 ; i < list.childNodes.length ; i++) {
		Patterns += list.childNodes[i].getAttribute("label");
		if (i + 1 != list.childNodes.length) // Do not put a space at the end.
			Patterns += " ";
	}
	return Patterns;
}

// Save the settings and close the window.
function saveSettings() {
	Patterns = getPatterns();

	prefObj = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService);

	var Branch = prefObj.getBranch("adblock.");
	Branch.setCharPref("patterns", Patterns);
	var checkbox = document.getElementById("enabled");
	Branch.setBoolPref("enabled", checkbox.checked);

	var radiohide = document.getElementById("radio-hide");
	Branch.setBoolPref("hide", radiohide.getAttribute("selected") == "true");
	
	prefObj.savePrefFile(null); // save the prefs to disk 
	persistListIndex(); // persists the current listitem's index in a separate listbox-attribute
	window.close(); // close out.
}

function about() {
	window.open("http://adblock.mozdev.org/","About"); 
}

// DEBUG_FUNCTIONS:
function listArray(ar) {
	str = "content-size: " + ar.length + "\n";
	for (var i = 0 ; i < ar.length; i++) {
		str += "-"+ar[i]+"-\n";
	}
	
	return str;
}

// lists everything in an object -- Useful debug-function.
function listObject(obj, s) {
	var res = "List: " + obj + "\n";
	for(var list in obj) {		
		if (list.indexOf(s) != -1)
			res += list + ", ";
	}
		
	return res;
}

